
#INCLUDE "common.ch"
#INCLUDE "inkey.ch"
#INCLUDE "appevent.ch"
#INCLUDE "xbp.ch"
#INCLUDE "font.ch"
#INCLUDE "dcmsg.CH"
#INCLUDE "dcdialog.CH"
#INCLUDE "nls.CH"

//////////////////////////////////////////////////////////////////////
//
//  DC_XBPGET.PRG
//
//  Copyright:
//     Alaska Software Inc., (c) 1997. All rights reserved.
//     Modified by DONNAY Software
//
//  Contents:
//     XbpGet implementation.
//
//  Remarks:
//     Sample implementation of a XbpGet class derived from XbpSle which uses
//     a overloaded TextMode Get object as server for the picture and buffer
//     management.
//     This class demonstrates the use of the XbpSle:AutoKeyboard flag
//     and the use of the GET class.
//
//     Note:
//     When validation is necessary, please refer to the validation
//     mechanisms of the DataRef class.
//
//  Return:
//     returns 0 always
//
//////////////////////////////////////////////////////////////////////

#ifndef xbeM_Enter
  #define xbeM_Enter  (005 + xbeB_Event)
#endif

#ifndef xbeM_Leave
  #define xbeM_Leave  (006 + xbeB_Event)
#endif


* **********************************************************************
* InvisibleGet service class
* **********************************************************************
CLASS InvisibleGet FROM Get
   EXPORTED:
   METHOD init
   METHOD display
ENDCLASS

METHOD InvisibleGet:Init( nRow, nCol, bVarBlock, cVarName, cPicture, ;
                          cColor, bValid, bWhen )
Grok(cColor,nRow,nCol)
RETURN ::Sle:Init( bVarBlock, cVarName, cPicture, bValid, bWhen )


METHOD InvisibleGet:display()
RETURN self



* **********************************************************************
* XbpGet class
* **********************************************************************
CLASS DC_XbpGet FROM DC_XbpGetSuper, DC_XbaseParts

   EXPORTED:
   VAR    Get                     // the imbeedded invisible service GET
   VAR    Picture                 // my picture
   VAR    pictureBlock            // picture code block
   VAR    exitState               // exit state (compatability)
   VAR    IsProtected             // protected flag (compatability)
   VAR    Buffer                  // GET buffer (compatability)
   VAR    row                     // display row (compatability)
   VAR    col                     // display col (compatability)
   VAR    options                 // GET options
   VAR    type                    // GET type
   VAR    original                // GET original value
   VAR    confirm                 // Confirm input with ENTER key
   VAR    password                // GET is a password
   VAR    proper                  // GET is proper case
   VAR    range                   // Range of values allowed
   VAR    passbuffer              // Password buffer
   VAR    oldPicture
   VAR    NextObject              // Object to pass focus to
   VAR    keyblock                // code block to evaluate after each key
   VAR    value                   // The current value during editing
                                  ** overloaded from XbpSle
   VAR color
   VAR varBlock
   VAR optionArray
   VAR popupFlag       // cargo[9]
   VAR popupContainer  // cargo[4]
   VAR popupButton     // cargo[7]
   VAR comboOptions    // cargo[8]
   VAR comboData
   VAR comboHeight
   VAR comboWidth
   VAR comboField
   VAR comboReturn
   VAR comboFont
   VAR comboPres
   VAR sayObject
   VAR comboStatic
   VAR lastKey
   // VAR origSize
   VAR dateFlag        // flag for date fields when using '@K' picture
   VAR comboFlag
   VAR comboKeyDrop    // flag to drop down box on first key
   VAR dataAlias
   VAR popupWhen
   VAR popupHide

   // for calculator-style input of numeric values
   CLASS VAR decimal  SHARED
   CLASS VAR thousand SHARED
   VAR isFloat
   VAR isMinus
   VAR typeOut
   VAR number
   VAR isCalc
   VAR varName
   VAR template
   VAR heightPadding
   VAR isGetDataOnLostFocus
   VAR isValidateOnChange
   VAR passwordChar

   CLASS METHOD InitClass
   METHOD typeOut
   METHOD LbDown

   METHOD Init, Create            // Lifecycle
   METHOD Keyboard                // overloaded keyboard handling
   METHOD LbUp                    // to inquire mark
   METHOD RbUp                    // to fix cut and paste bug

   METHOD CutMarked
   METHOD PasteMarked
   METHOD DeleteMarked            // manipulate mark
                                  ** overloaded from DataRef
   METHOD SetData, GetData        // overloaded nSetData, GetData
   METHOD Undo                    // overloaded Undo to Get
   METHOD toDecPos
   METHOD Clear
   METHOD Home

   METHOD SetEditBuffer           // change edit buffer
   METHOD _Assign                 // Clipper GET compatability

   METHOD GetSetFocus             // Fix problem with :dectopos in some picture clauses
   METHOD GetKillFocus            // same as above

   METHOD Refresh
   METHOD WhenHideEval
   METHOD HideShow

   METHOD DragDrop

   METHOD SetInputFocus

   METHOD MakePicture
   METHOD Destroy

   METHOD ResizeSetSize

ENDCLASS

/*
* Store country-specific delimiters in class variables
*/
CLASS METHOD DC_XbpGet:initClass
   ::decimal  := SetLocale( NLS_SDECIMAL  )
   ::thousand := SetLocale( NLS_STHOUSAND )
RETURN self


* **********************************************************************
* Init method
* **********************************************************************
METHOD DC_XbpGet:Init( oParent, oOwner, aPos, aSize, aPP, lVisible, oGetList )

LOCAL aArray, xOptions, bBlock, aGetList, aGetListItem, aGetOptions

DEFAULT aPP TO {}

* initialisation of base class
AAdd( aPP, { XBP_PP_COMPOUNDNAME, FONT_DEFFIXED_SMALL } )
::XbpSle:init( oParent, oOwner, aPos, aSize, aPP, lVisible )

::password := .f.
::confirm := .t.
::isCalc := .F.
::typeOut := .F.
::isFloat := .F.
::isMinus := .F.
::picture := ''
::dateFlag := .T.
::comboFlag := .F.
::comboKeyDrop := .F.
::heightPadding := DC_XbpGetHeightPadding()
::isGetDataOnLostFocus := .t.
::isValidateOnChange := DC_GetValidateOnChange()

::DC_XbaseParts:init( oGetList,,self )
aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]
aGetOptions := ::getList:getOptions

::template := aGetOptions[cGETOPT_GETTEMPLATE]

/*
IF Empty(::template)
  ::template := 'W'
ENDIF
*/

IF Valtype(aGetlistItem[nGETLIST_POINTER]) = 'N'
  ::sayObject := aGetList[aGetlistItem[nGETLIST_POINTER],oGETLIST_OBJECT]
ENDIF

xOptions := aGetListItem[xGETLIST_OPTIONS]
::optionArray := xOptions

IF Valtype(aGetListItem[xGETLIST_OPTIONS3]) = 'L'
  ::unreadable := aGetListItem[xGETLIST_OPTIONS3]
ENDIF

::clipSiblings := .t.
::clipChildren := .t.
IF Valtype(aGetListItem[cGETLIST_PICTURE]) == 'B'
  ::pictureBlock := aGetListItem[cGETLIST_PICTURE]
  ::picture := Eval(::pictureBlock,self)
ELSE
  ::picture := aGetListItem[cGETLIST_PICTURE]
ENDIF
::row := aPos[2]
::col := aPos[1]

::popupFlag := .f.
IF !Empty(aGetListItem[bGETLIST_WHEN])
  ::whenBlock := DC_GetBlock(aGetListItem[bGETLIST_WHEN])
ENDIF
IF !Empty(aGetListItem[bGETLIST_HIDE])
  ::hideBlock := DC_GetBlock(aGetListItem[bGETLIST_HIDE])
ENDIF
IF !Empty(aGetListItem[bGETLIST_PROTECT])
  ::protectBlock := DC_GetBlock(aGetListItem[bGETLIST_PROTECT])
ENDIF

::cargo := { ::getListPointer, aGetList, aGetListItem[xGETLIST_CARGO] }

::lbDown := DC_XbpGetLbDown()
::lbClick := DC_XbpGetLbClick()

IF Empty(aGetListItem[nGETLIST_SUBTYPE]) // Not being used as default Cell Editor
  DC_ColorGets( self,2,aGetOptions[aGETOPT_COLORGETS])
  ::killInputFocus := {|x,y,o|  ;
                        IIF(::isGetDataOnLostFocus,o:getData(),nil), ;
                        IIF(::isGetDataOnLostFocus,o:setData(),nil), ;
                        o:home(), ;
                        DC_HiliteGets(o,2,o:getList:getOptions[nGETOPT_HILITECOLOR]), ;
                        DC_ColorGets(o,2,o:getList:getOptions[aGETOPT_COLORGETS]), ;
                        _PopupButtonShow(o,.f.) }
  ::setInputFocus := {|x,y,o| ;
                       o:DateFlag := .T., ;
                       o:Home(), ;
                       DC_HiliteGets(o,1,o:getList:getOptions[nGETOPT_HILITECOLOR]), ;
                       DC_ColorGets(o,1,o:getList:getOptions[aGETOPT_COLORGETS]), ;
                       _PopupButtonShow(o,.t.)}
ENDIF

IF !Empty(aGetListItem[bGETLIST_VALID])
  bBlock := DC_GetBlock(aGetListItem[bGETLIST_VALID])
  ::validate := bBlock
ENDIF
::bufferLength := 1000
IF Valtype(aGetListItem[xGETLIST_OPTIONS4]) # 'L'
  aGetListItem[xGETLIST_OPTIONS4] := !aGetOptions[lGETOPT_CONFIRM]
ENDIF

::confirm := !aGetListItem[xGETLIST_OPTIONS4]
::passwordChar := '*'
IF Valtype(aGetListItem[xGETLIST_OPTIONS5]) = 'L'
  ::password := aGetListItem[xGETLIST_OPTIONS5]
ELSEIF Valtype(aGetListItem[xGETLIST_OPTIONS5]) = 'A'
  DEFAULT aGetListItem[xGETLIST_OPTIONS5] := {}
  ASize(aGetListItem[xGETLIST_OPTIONS5],2)
  IF !Empty(aGetListItem[xGETLIST_OPTIONS5,2])
    aGetListItem[xGETLIST_OPTIONS5,1] := .t.
  ENDIF
  DEFAULT aGetListItem[xGETLIST_OPTIONS5,2] := '*'
  ::password := aGetListItem[xGETLIST_OPTIONS5,1]
  ::passwordChar := aGetListItem[xGETLIST_OPTIONS5,2]
ENDIF

aArray := aGetListItem[xGETLIST_OPTIONS6]
IF Valtype(aArray) == 'L'
  ::proper := {.F.,' '}   // maintain current function
ELSEIF Valtype(aArray) == 'A'
  ASize(aArray,2)
  DEFAULT aArray[1] := .t.
  DEFAULT aArray[2] := {}
  ASize(aArray[2],2)
  DEFAULT aArray[2,1] := .f.
  DEFAULT aArray[2,2] := ' '
  IF aArray[1]
    ::proper := aArray[2]
  ENDIF
ENDIF

IF Valtype(aGetListItem[xGETLIST_OPTIONS8]) = 'A'
  ::range := aGetListItem[xGETLIST_OPTIONS8]
ENDIF

::keyBlock := aGetListItem[xGETLIST_OPTIONS9]
::color := aGetListItem[aGETLIST_COLOR]

IF !(::color == NIL)

  DC_XbpSetColor( ::color, self, oParent, ;
                  aGetListItem[aGETLIST_PRESENTATION], .t., .f. )
ENDIF

IF !Empty(aGetListItem[bGETLIST_GROUP])
  bBlock := DC_GetBlock(aGetListItem[bGETLIST_GROUP],'O')
  IF Valtype(bBlock) = 'B'
    Eval( bBlock, self )
  ENDIF
ENDIF

IF !Empty(aGetListItem[bGETLIST_PREEVAL])
  bBlock := DC_GetBlock(aGetListItem[bGETLIST_PREEVAL])
  IF Valtype(bBlock) = 'B'
    Eval( bBlock, self )
  ENDIF
ENDIF

::varName := aGetListItem[cGETLIST_VARNAME]
IF Valtype(::varName) == 'A'
  ::varName := ::varName[1]
ENDIF

RETURN self

* **********************************************************************
* Create method
* **********************************************************************
METHOD DC_XbpGet:Create( oParent, oOwner, aPos, aSize, aPP, lVisible )

LOCAL aArray, nWidth, nHeight, bBlock, j, nStartRow, nStartCol, cCaption, ;
      xOptions, lPopValid, xData, aGetListItem, aGetOptions, ;
      nHotKey, oFont, cFont, cNominalPointSize, cFamilyName, bCombo

aGetListItem := ::getList:getListArray[::getListPointer]
aGetOptions := ::getList:getOptions

DEFAULT aGetListItem[xGETLIST_OPTIONS2] := {}
ASize(aGetListItem[xGETLIST_OPTIONS2],16)
::isCalc := aGetListItem[xGETLIST_OPTIONS2,14]
::popupWhen := aGetListItem[xGETLIST_OPTIONS2,15]
::popupHide := aGetListItem[xGETLIST_OPTIONS2,16]
DEFAULT ::isCalc := .F.

::datalink := DC_GetBlock(aGetListItem[bGETLIST_VAR])

* now the datalink must have been specified
IF ! ValType( ::DataLink ) == "B"
  ::Status := XBP_STAT_FAILURE
  RETURN self
ENDIF

IF ::isCalc // .OR. Valtype(Eval(::dataLink)) == 'N'
  ::align := XBPSLE_RIGHT
ENDIF

IF Valtype(aGetListItem[cGETLIST_PICTURE]) == 'B'
  ::pictureBlock := aGetListItem[cGETLIST_PICTURE]
  ::picture := Eval(::pictureBlock,self)
ELSE
  ::picture := aGetListItem[cGETLIST_PICTURE]
ENDIF

* create the SLE and disable keyboard handling
::XbpSle:autoKeyboard := .F.

IF Valtype(aGetListItem[cGETLIST_FONT]) == 'O'
  ::SetFont(aGetListItem[cGETLIST_FONT])
ELSEIF !Empty(aGetListItem[cGETLIST_FONT])
  IF !Upper(aGetListItem[cGETLIST_FONT]) == 'DEFAULT'
    ::SetFontCompoundName( Alltrim(aGetListItem[cGETLIST_FONT]) )
  ENDIF
ELSEIF !Empty(::getList:getFont)
  IF Valtype(::getList:getFont) == 'O'
    ::SetFont(::getList:getFont)
  ELSE
    aGetListItem[cGETLIST_FONT] := ::getList:getFont
    ::SetFontCompoundName( Alltrim(::getList:getFont) )
  ENDIF
ELSEIF ::getList:scaleFactor[5]
  ::setFontCompoundName( "10.Courier New" )
ELSE
  ::setFontCompoundName( "10.Courier" )
ENDIF

::XbpSle:Create( oParent, oOwner, aPos, aSize, aPP, lVisible )

* now initialize the get object
::Get := InvisibleGet():New(,, ::Datalink, "", ::Picture )
::Get:SetFocus()
::BufferLength := Len( ::Get:buffer )
::exitState := 0
::IsProtected := .f.
::row := 0
::col := 0
::passbuffer := ''
::type := ::Get:Type
::XbpSle:rbUp := {|aPos|::RbUp(aPos)}

xOptions := ::optionArray

IF Valtype(xOptions) = 'A' .AND. !Empty(xOptions) .AND. xOptions[1,1,1] == 106
  IF IsMemberVar(::xbpSle,'cueBanner')
    ::xbpSle:setCueBanner(xOptions[1,1,2])
  ENDIF
ENDIF

IF Valtype(::hideBlock)='B' .AND. Eval(::hideBlock)
  ::hide()
ENDIF

nWidth := ::currentSize()[1]
nHeight := ::currentSize()[2]
nStartCol := ::currentPos()[1]
nStartRow := ::currentPos()[2]

IF ::GetList:scaleFactor[5]
  oFont := ::setFont()
  IF Valtype(oFont) = 'O'
    cFont := Alltrim(Str(Round(oFont:nominalPointSize * ::GetList:scaleFactor[4],0))) + ;
               '.' + oFont:familyName
    ::setFontCompoundName(cFont)
  ENDIF
ENDIF

IF nWidth = 0
  xData := Eval(::dataLink,,self)
  cCaption := DC_XtoC(xData)
  IF !Empty(aGetListItem[cGETLIST_PICTURE])
    IF Valtype(aGetListItem[cGETLIST_PICTURE]) == 'B'
      ::pictureBlock := aGetListItem[cGETLIST_PICTURE]
      cCaption := TransForm(xData,Eval(aGetListItem[cGETLIST_PICTURE]))
    ELSE
      cCaption := TransForm(xData,aGetListItem[cGETLIST_PICTURE])
    ENDIF
  ELSEIF Valtype(xData) = 'D'
    cCaption := DtoC(xData)
  ELSEIF Valtype(xData) = 'N'
    cCaption := Str(xData)
  ENDIF

  IF !Empty(::template)
    aArray := GraQueryTextBox( ::lockPS(), Repl(::template,Len(cCaption) ))
  ELSE
    aArray := GraQueryTextBox( ::lockPS(), cCaption )
  ENDIF

  IF !Empty(aGetListItem[cGETLIST_FONT])
    nWidth := (aArray[3,1] - aArray[1,1]) + 6
    nHeight := (aArray[1,2] - aArray[2,2]) + 4 + DC_XbpGetHeightPadding()
  ELSE
    nWidth := Max(nWidth,(aArray[3,1] - aArray[1,1]) + 6)
    nHeight := Max(nHeight,(aArray[1,2] - aArray[2,2]) + 2)
  ENDIF

  ::unlockPS()
  ::setSize( {nWidth, nHeight })

ENDIF

::SetData()
// ::origSize := ::currentSize()

lPopValid := aGetListItem[xGETLIST_OPTIONS2,5]
DEFAULT lPopValid := .f.

bBlock := DC_GetBlock(aGetListItem[xGETLIST_OPTIONS2,1])

// If no POPUP block, check if POPVALID is a logical TRUE
// and there is a VALID block.  Use it as the POPUP block

IF Valtype(bBlock) # 'B' .AND. lPopValid .AND. Valtype(::validate) = 'B'
  bBlock := ::validate
ENDIF

// There is a POPUP block or a COMBO
IF Valtype(bBlock) = 'B' .OR. ;
   (Valtype(xOptions) = 'A' .AND. !Empty(xOptions) .AND. xOptions[1,1,1] == 120)
  ::popupContainer := { PICKPOINTER, LEN(::getList:editControls) }

  DEFAULT oParent := ::setParent()
  nStartCol += ( nWidth )
  cCaption := aGetListItem[xGETLIST_OPTIONS2,6]
  aArray := aGetListItem[xGETLIST_OPTIONS2,11] // PopUp Hot-Key
  IF Valtype( aArray ) $ 'AN'
    IF Valtype( aArray ) = 'N'
      aArray := { aArray }
    ENDIF
    FOR j := 1 TO Len(aArray)
      _AddHotKey( aArray[j], ;
                  _PopKey(self), ;
                  ::whenBlock, ;
                  ::hideBlock, ;
                  IIF( aGetListItem[xGETLIST_OPTIONS2,12],nil,self), ;
                  ::getList:keyboardControls, ;
                  self )
    NEXT
  ENDIF

  IF Empty(cCaption) .AND. aGetListItem[xGETLIST_OPTIONS2,10] == NIL
    cCaption := DC_GetPopupCaption()
    ASize(cCaption,6)
    // [1]  Font
    // [2]  Caption
    // [3]  Type DCGUI_POPUPSTYLE_OUTSIDE or DCGUI_POPUPSTYLE_IMBEDDED
    // [4]  Width
    // [5]  Height
    // [6]  Hot-Key
  ELSE
    DEFAULT aGetListItem[xGETLIST_OPTIONS2,10] := DCGUI_POPUPSTYLE_OUTSIDE
    DEFAULT aGetListItem[xGETLIST_OPTIONS2,12] := .f.
    IF Empty(aGetListItem[xGETLIST_OPTIONS2,6])
      aGetListItem[xGETLIST_OPTIONS2,6] := 'u'
      DEFAULT aGetListItem[xGETLIST_OPTIONS2,7] := '10.Marlett'
    ENDIF
    cCaption := { aGetListItem[xGETLIST_OPTIONS2,7], ; // font
                  aGetListItem[xGETLIST_OPTIONS2,6], ; // caption
                  aGetListItem[xGETLIST_OPTIONS2,10],; // style
                  nil,                               ; // width
                  nil,                               ; // height
                  nil }                                // hot-key
  ENDIF

  IF Valtype(cCaption[5])='N'
    nHeight := cCaption[5]
  ENDIF
  IF Valtype(cCaption[4])='N'
    nWidth := cCaption[4]
  ELSE
    nWidth := nHeight
  ENDIF
  IF Valtype(aGetListItem[xGETLIST_OPTIONS2,8]) = 'N'
    nWidth := aGetListItem[xGETLIST_OPTIONS2,8]
  ENDIF
  IF Valtype(aGetListItem[xGETLIST_OPTIONS2,9]) = 'N'
    nHeight := aGetListItem[xGETLIST_OPTIONS2,9]
  ENDIF

  IF Valtype(cCaption[3])='N' .AND. cCaption[3] = DCGUI_POPUPSTYLE_IMBEDDED
    nHeight -= 4
    nWidth -= 4
    ::setSize({::currentSize()[1]+nWidth,::currentSize()[2]})
    oParent := self
    nStartCol := ::currentSize()[1] - nWidth - 5
    nStartRow := 0
  ENDIF

  IF Valtype(DC_PopupButtonStyle()) == 'O'
    #if XPPVER > 1900000
    ::popupButton := DC_XbpPushButtonXP():new( oParent, , ;
             { nStartCol+1, nStartRow }, { nWidth, nHeight }, nil, lVisible, ::getList, DC_PopupButtonStyle(),.t. )
    #endif
  ELSE
    ::popupButton := XbpPushButton():new( oParent, , ;
             { nStartCol+1, nStartRow }, { nWidth, nHeight }, nil, lVisible )
  ENDIF

  IF Valtype(cCaption[1]) = 'C'
    cFont := cCaption[1]
    IF ::GetList:scaleFactor[5]
      cNominalPointSize := Alltrim(Str(Round(::GetList:scaleFactor[3]*Val(cFont),0)))
      cFamilyName := Substr(cFont,At('.',cFont))
      cFont := cNominalPointSize + cFamilyName
    ENDIF
    ::popupButton:SetFontCompoundName(cFont)
  ENDIF
  IF Valtype(cCaption[2])='A'
    ::popupButton:caption := DC_GetBitmap(cCaption[2,1])
  ELSE
    ::popupButton:caption := DC_GetBitmap(cCaption[2])
  ENDIF
  ::popupButton:clipSiblings := .T.
  ::popupButton:clipParent := .t.
  ::popupButton:create(,,,,,,::popupButton:caption,cFont)

  IF DC_GetPopupAutoHide()
    ::popupButton:hide()
  ELSEIF Valtype(::popupHide) = 'B' .AND. Eval(::popupHide,self)
    ::popupButton:hide()
  ENDIF
  IF Valtype(::popupWhen) = 'B' .AND. !Eval(::popupWhen,self)
    ::popupButton:disable()
  ENDIF
  ::popupButton:killInputFocus := ;
      {|a,b,o,n|n:=NextAppEvent(@a,@b,@o), ;
                IIF(o==self,nil,PostAppEvent(xbeP_KillInputFocus,,,self))}

  ::comboOptions := xOptions

  IF Valtype(xOptions) == 'A'
    ASize( xOptions[1,1], 12 )
    ::comboHeight := xOptions[1,1,2]
    ::comboWidth := xOptions[1,1,6]
    ::comboData := xOptions[1,1,3]
    ::comboField := xOptions[1,1,4]
    ::comboReturn := xOptions[1,1,5]
    ::comboFont := xOptions[1,1,10]
    ::comboPres := xOptions[1,1,11]
    ::comboKeyDrop := xOptions[1,1,12]
    DEFAULT ::comboKeyDrop := .f.
  ENDIF

  IF Valtype(xOptions) = 'A' .AND. !Empty(xOptions) .AND. xOptions[1,1,1] == 120
    xOptions := xOptions[1,1]
    ASize(xOptions,11)
    IF !Empty(xOptions[7])
      ::popupButton:caption := xOptions[7]
      ::popupButton:configure()
    ENDIF
    IF !Empty(xOptions[8])
      IF Valtype(xOptions[8]) == 'C'
        ::popupButton:setFontCompoundName(xOptions[8])
      ELSEIF Valtype(xOptions[8]) == 'O'
        ::popupButton:setFont(xOptions[8])
      ENDIF
    ENDIF
    nHotKey := xOptions[9]
    IF Empty(nHotKey)
      nHotKey := cCaption[6]
    ENDIF
    IF !Empty(nHotKey)
      ::keyboard := { |a,b,o|IIF(a=nHotKey,Eval(::popupButton:activate),nil) }
    ENDIF
    bCombo := _GetCombo(self,aGetListItem[lGETLIST_PIXEL],aGetOptions,cCaption[3])
    bBlock := _PopClick( bCombo, self, nil, lPopValid, ::getList )
    IF ::comboKeyDrop
      ::keyboard := DC_MergeBlocks(::keyboard,{|nKey,a,oGet|_GetComboKeyboard(bCombo,nKey,oGet)})
    ENDIF
  ELSE

    bBlock := _PopClick( bBlock, self, ;
              DC_GetBlock(aGetListItem[bGETLIST_READVAR]), lPopValid, ::getList )
  ENDIF
  ::popupButton:cargo := ;
      { { PICKPOINTER, ::getListPointer, self }, cCaption[2], .t., bBlock, ;
          IF(ValType(aGetListItem[bGETLIST_PROTECT])=='B',;
          aGetListItem[bGETLIST_PROTECT],aGetOptions[bGETOPT_EDITPROTECT]),;
          aGetListItem[xGETLIST_OPTIONS2,13] }
  ::popupButton:activate := {|a,b,o|_PopUpButtonActivate(a,b,::popupButton)}
  IF Len(aGetListItem[xGETLIST_OPTIONS2]) < 3
    ASize(aGetListItem[xGETLIST_OPTIONS2],3)
  ENDIF
  IF Valtype(aGetListItem[xGETLIST_OPTIONS2,3]) = 'L'
    ::popupButton:tabstop := aGetListItem[xGETLIST_OPTIONS2,3]
  ELSEIF Valtype(aGetListItem[lGETLIST_TABSTOP]) = 'L'
    ::popupButton:tabstop := aGetListItem[lGETLIST_TABSTOP]
  ELSE
    ::popupButton:tabstop := .f.
  ENDIF
  ::popupButton:keyboard := { |a,b,o|IIF(a=xbeK_ENTER,Eval(o:activate),nil) }
  AADD(::getList:pickControls,::popupButton)

ENDIF

::DC_XbaseParts:setOrig(self)

RETURN self

* -----------------

METHOD DC_XbpGet:destroy()

::xbpSle:destroy()
IF Valtype(::popupButton) == 'O' .AND. ::popupButton:status() > 0
  ::popupButton:destroy()
ENDIF
IF Valtype(::comboStatic) == 'O' .AND. ::comboStatic:status() > 0
  ::comboStatic:destroy()
ENDIF

::picture := ;
::number  := ;
::isMinus := ;
::isFloat := ;
::isCalc  := ;
::typeOut := NIL

RETURN self

* -----------------

METHOD DC_XbpGet:refresh()

IF !(::get:varget() == ::get:unTransform()) .OR. !Empty(::pictureBlock)
  ::setData()
ENDIF

DC_XbpSetColor( ::color, self, nil, nil, .t., .f. )

RETURN self

* -----------------

METHOD DC_XbpGet:setInputFocus( mp1, mp2 )

LOCAL cPicture, lSetMarked := .f.

IF ::status() == 0
  RETURN self
ENDIF

::get:setFocus()
::xbpSle:setInputFocus( mp1, mp2 )

IF ::isCalc
  ::setEditbuffer()
//  ::get:picture := ::picture
ELSE
  IF Valtype(::pictureBlock) == 'B'
    cPicture := Eval(::pictureBlock,self)
    IF Valtype(cPicture) == 'C'
      ::get:picture := cPicture
      PostAppEvent(xbeP_Keyboard,xbeK_HOME,,self)
    ENDIF
  ENDIF
  IF Empty(cPicture)
    cPicture := ::get:picture
  ENDIF
  IF !Empty(cPicture) .AND. '@'$cPicture .AND. 'K'$Upper(cPicture)
    lSetMarked := .t.
  ENDIF
  IF !(::get:changed) .AND. lSetMarked
    ::setMarked({1,Len(Trim(::get:buffer))+1})
  ENDIF
ENDIF

RETURN self

* **********************************************************************
* Overloaded SetData()
*
* This method should not behave like the normal XbpSle:SetData() method
* which needs strings. This method will rather call the codeblock and
* use its return value and convert this value to the given datatype.
* **********************************************************************
METHOD DC_XbpGet:SetData( xValue )

   * call updatebuffer on GET, this will call the datalink and set
   * the :Get:buffer value

   LOCAL cPicture, nNumber

   Grok(xValue)

   IF PCount() == 1
      RETURN ::_Assign(xValue)
   ENDIF

   ::exitState := 0
   ::get:exitState := 0

   IF ::bufferLength == 0
     ::bufferLength := Len(::get:buffer)
   ENDIF

   IF ::isCalc
     IF PCount() == 0
        nNumber := Eval( ::dataLink,,self )
     ENDIF
     ::number  := nNumber
     ::makePicture()
     ::setEditBuffer()
     RETURN ::number
   ENDIF

   IF Valtype(::pictureBlock) == 'B'
     cPicture := Eval(::pictureBlock,self)
     IF Valtype(cPicture) == 'C'
       ::get:picture := cPicture
     ENDIF
   ENDIF

   ::Get:UpdateBuffer()

   IF ::password
     IF !Empty(::Get:Buffer)
       ::passBuffer := ::Get:Buffer
     ELSE
       ::passBuffer := ''
     ENDIF
     ::Get:Buffer := Pad(Repl(::passwordChar,Len(Alltrim(::passBuffer))),::bufferLength)
   ENDIF

   ::Buffer := ::Get:Buffer

   IF ::original == nil
      ::original := ::Buffer
   ENDIF

   * now set the data in the SLE class
   ::SetEditBuffer( ::Get:Buffer )

   * now set the initial mark
   IF Set( _SET_INSERT )
      ::SetMarked ( { ::Get:Pos, ::Get:Pos } )
   ELSE
      ::SetMarked ( { ::Get:Pos, ::Get:Pos + 1 } )
   ENDIF

RETURN self

* ------------
* Delegate the xbeM_LbDown event to the parent. This disables the
* marking of digits in the entry field.

METHOD DC_XbpGet:lbDown( mp1, mp2 )

IF ::isCalc
  ::setParent():lbDown( mp1, mp2 )
ELSE
  ::xbpSLE:lbDown( mp1, mp2 )
ENDIF

RETURN self

* --------------

METHOD DC_XbpGet:dragDrop( x, y )

RETURN self

* --------------
* Input focus is obtained when the left mouse button is released.
METHOD DC_XbpGet:lbUp( mp1, mp2 )

LOCAL aMarked

IF ::isCalc
   SetAppFocus( self )
ELSE
   aMarked := ::queryMarked()
   * now locate Get on the specified position
   ::Get:Pos := aMarked[1]
   * now set the mark when insert mode is not active
   IF aMarked[2] - aMarked[1] == 0 .AND. ! Set( _SET_INSERT )
      ::SetMarked( { aMarked[1], aMarked[1]+1 } )
   ENDIF
ENDIF

RETURN ::XbpSle:LbUp( mp1, mp2 )


* Suppress the default popup menu of the XbpSle class
* by dirceting the xbeM_RbUp event to the parent.
METHOD DC_XbpGet:rbUp( mp1, mp2 )

IF ::isCalc
  ::setParent():rbUp( mp1, mp2 )
ELSE
  ::Get:Buffer := ::EditBuffer()
  ::xbpSLE:rbUp( mp1, mp2 )
ENDIF

RETURN self


* **********************************************************************
* Overloaded GetData()
*
* Return untransformed contents of the GET buffer
* **********************************************************************
METHOD DC_XbpGet:GetData()

   LOCAL xValue

   IF ::status() == 0
     RETURN xValue
   ENDIF

   IF ::isCalc
     IF Valtype( ::dataLink ) == "B"
       Eval( ::dataLink, ::number, self )
     ENDIF
     RETURN ::number
   ENDIF

   IF ::get:type == 'N' .AND. Val(::get:buffer) > 0 .AND. ;
         At( " ", Alltrim( ::Get:buffer)) != 0 .AND. ;
         Empty(Left(::Get:buffer,1))

     ::Get:buffer := TransForm( Val( StrTran(Strtran(::Get:buffer,' ',''),',','')), ::Get:picture)
     ::SetEditBuffer(::Get:buffer)
   ENDIF

   IF ::password
      xValue := ::passbuffer
   ELSE
      xValue := ::Get:Untransform()
   ENDIF

   * store value to datalink

   Eval( ::DataLink, xValue, self )

RETURN xValue

* **********************************************************************

/*
METHOD DC_XbpGet:toDecPos()

LOCAL nDecPos, i, cPrefix, lMinus := '-'$::get:buffer, cDecChar

IF ::Get:Type = 'N'
  cDecChar := ::decimal // SetLocale( NLS_SDECIMAL)
  IF Valtype(::picture) = 'C' .AND. '@ZE' $ ::picture
    nDecPos := At(cDecChar,Alltrim(Strtran(::picture,'@ZE','')))
  ELSEIF Valtype(::picture) = 'C' .AND. '@Z' $ ::picture
    nDecPos := At(cDecChar,Alltrim(Strtran(::picture,'@Z','')))
  ELSEIF Valtype(::picture) = 'C' .AND. '@$ ' $ ::picture
    nDecPos := At(cDecChar,Alltrim(Strtran(::picture,'@$ ','')))+2
  ELSE
    nDecPos := At(cDecChar,::get:Buffer)
  ENDIF

  IF nDecPos > 0 .OR. .t.
    IF nDecPos > 0 .AND. nDecPos < 5 .AND. lMinus // fix bug in :toDecPos() method of Get()
      ::Get:home()
      FOR i := 1 TO nDecPos-1
        ::Get:right()
      NEXT
      ::get:Buffer := StrTran(::get:buffer,'- ','-0')
      cPrefix := Alltrim(Substr(::get:buffer,1,nDecPos-1))
      ::get:Buffer := PadL(cPrefix,nDecPos-1)+Substr(::get:Buffer,nDecPos)
    ELSE
      // ::Get:home()
      // FOR i := 1 TO nDecPos-1
      //  ::Get:right()
      // NEXT
      // cPrefix := Alltrim(Substr(::get:buffer,1,nDecPos-1))
      // ::get:Buffer := PadL(cPrefix,nDecPos-1)+Substr(::get:Buffer,nDecPos)

      // ::get:condClear()
      // ::get:toDecPos()

      IF Empty(::get:Buffer) // fix bug in :toDecPos() method of Get()
        ::get:Buffer := Stuff(::get:Buffer,nDecPos,1,cDecChar)
        IF lMinus
          ::get:Buffer := Stuff(::get:Buffer,nDecPos-1,1,'-')
        ENDIF
      ENDIF

    ENDIF
  ENDIF
ENDIF
RETURN nil
*/

METHOD DC_XbpGet:toDecPos()

LOCAL nDecPos, i, cPrefix, lMinus := '-'$::get:buffer, cDecChar

IF ::Get:Type = 'N'
  cDecChar := SetLocale( NLS_SDECIMAL)

  IF Valtype(::picture) = 'C' .AND. '@Z' $ ::picture
    nDecPos := At(cDecChar,Alltrim(Strtran(::picture,'@Z','')))
  ELSEIF Valtype(::picture) = 'C' .AND. '@$ ' $ ::picture
    nDecPos := At(cDecChar,Alltrim(Strtran(::picture,'@$ ','')))+2
  ELSE
    nDecPos := At(cDecChar,::get:Buffer)
  ENDIF

  IF nDecPos > 0 .OR. .t.
    IF nDecPos > 0 .AND. nDecPos < 5 .AND. lMinus // fix bug in :toDecPos() method of Get()
      ::Get:home()
      FOR i := 1 TO nDecPos-1
        ::Get:right()
      NEXT
      ::get:Buffer := StrTran(::get:buffer,'- ','-0')
      cPrefix := Alltrim(Substr(::get:buffer,1,nDecPos-1))
      ::get:Buffer := PadL(cPrefix,nDecPos-1)+Substr(::get:Buffer,nDecPos)
    ELSE
      // ::Get:home()
      // FOR i := 1 TO nDecPos-1
      //   ::Get:right()
      // NEXT
      // cPrefix := Alltrim(Substr(::get:buffer,1,nDecPos-1))
      // ::get:Buffer := PadL(cPrefix,nDecPos-1)+Substr(::get:Buffer,nDecPos)
      // ::get:condClear()
      // ::get:toDecPos()

      IF Empty(::get:Buffer) // fix bug in :toDecPos() method of Get()
        ::get:Buffer := Stuff(::get:Buffer,nDecPos,1,cDecChar)
        IF lMinus
          ::get:Buffer := Stuff(::get:Buffer,nDecPos-1,1,'-')
        ENDIF
      ENDIF

    ENDIF
  ENDIF
ENDIF
RETURN nil

* **********************************************************************
METHOD DC_XbpGet:Clear( lClear )

DEFAULT lClear := .t.

::Get:clear := lClear
RETURN nil

*** **********************************************************************
* Overloaded Undo()
* **********************************************************************
METHOD DC_XbpGet:Undo()

   ::Get:Undo()
   ::Get:Buffer := Transform( ::Get:Original, ::Get:Picture )
   ::Buffer := ::Get:Buffer

RETURN ::SetEditBuffer( ::Get:buffer )

* **********************************************************************
* Change the edit buffer of the XbpSle.
* **********************************************************************
METHOD DC_XbpGet:SetEditBuffer( cBuffer )

LOCAL nLen, cDecChar

IF PCount() > 0

   ::XbpSle:SetData( cBuffer )

ELSEIF ::isCalc // Calculator style input

//   cDecChar := ::decimal // SetLocale( NLS_SDECIMAL)

   cDecChar := SetLocale( NLS_SDECIMAL)

   IF Right( ::picture, 1 ) == cDecChar
      * Number with decimal point but no decimal digits.
      nLen    := Len( ::picture )
      cBuffer := Left( Transform( ::number, ::picture +"9" ), nLen )
   ELSE
      cBuffer := Transform( ::number, ::picture )
   ENDIF

   cBuffer := LTrim( cBuffer )

   IF ::number == 0 .AND. ::IsMinus
      * A negative number is being entered but the value is still zero.
      cBuffer := "-" + cBuffer
   ENDIF

   IF ::isFloat
      * Set the caret behind the last digit in the edit buffer.
      ::xbpSLE:setData( cBuffer )
      nLen := Len( ::editBuffer() ) + 1
   ELSE
      * Set the caret to the left of the decimal point.
      ::xbpSLE:setData( cBuffer + ::decimal )
      nLen := Len( ::editBuffer() )
   ENDIF

   ::setMarked( {nLen, nLen} )
   ::typeOut()

ENDIF

RETURN self

* **********************************************************************
* Assign() method is for compabibility with Clipper Gets
* **********************************************************************
METHOD DC_XbpGet:_Assign( xValue, bDataLink )

DEFAULT bDataLink := ::dataLink

Eval(bDatalink,xValue,,self)

DC_GetEditUpdated(.t.)
::home()

::Get:OverStrike( SubStr(::Get:buffer,1,1) )
::home()
::setData()
IF ::isValidateOnChange
  ::Get:changed := .t.
ENDIF

RETURN self

* **********************************************************************
* Overloaded Keyboard handling
* **********************************************************************

METHOD DC_XbpGet:Keyboard( nKey )

   LOCAL aMarked, cChar, lHandled := .T., cSavePict, cDecChar
   LOCAL setPos    := .T., nPos, i, cPicture, xPos, nDecPos
   LOCAL cChr := IIf( nKey > 0 .AND. nKey < 255, Chr(nKey), " " ), ;
         cBuffer, dDate

   STATIC nAccent

   ::lastKey := nKey
   aMarked := ::QueryMarked()

   IF ::password
      ::Get:Buffer := ::passbuffer
   ENDIF

   // Many thanks to Joo Machado de Sousa for the Portuguese Character Support
   IF (nkey == 126  .AND. (DC_LangSet() == DCLANG_PORTUGUESE .OR. DC_LangSet() == DCLANG_PORTUGUESE_BR)) ;
         .OR. nkey ==180 .OR. nKey == 131168 .OR. nKey == 131166
        nAccent := nkey
        return self
   ELSEIF nAccent != nil
        cChar := chr(nkey)
        if nAccent == 126 .and. upper(cChar)$'AO'
         nkey := asc(substr('', at(cChar,'aoAO'), 1))
        elseif nAccent == 131166 .and. upper(cChar)$'AE'
         nkey := asc(substr('', at(cChar,'aeAE'), 1))
        elseif nAccent == 180 .and. upper(cChar)$'AEIOU'
         nkey := asc(substr('', at(cChar,'aeiouAEIOU'), 1))
        elseif nAccent == 131168 .and. upper(cChar)$'AEIOU'
         nkey := asc(substr('', at(cChar,'aeiouAEIOU'), 1))
        endif
        nAccent := nil
   ENDIF

   DO CASE

   CASE nKey == xbeK_SH_RIGHT
        ::Get:Pos ++
        aMarked[2] := ::Get:Pos
        ::SetMarked( aMarked )
        setPos := .F.

   CASE nKey == xbeK_SH_LEFT
        ::Get:Pos --
        aMarked[1] := ::Get:Pos
        ::SetMarked( aMarked )
        setPos := .F.

   CASE nKey == xbeK_SH_END
        aMarked[2] := ::Bufferlength + 1
        ::Get:Pos  := ::Bufferlength
        ::SetMarked( aMarked )
        setPos := .F.

   CASE nKey == xbeK_SH_HOME
        ::Get:Pos  := 1
        aMarked[1] := 1
        ::SetMarked( aMarked )
        setPos := .F.

   CASE nKey == xbeK_CTRL_U
        ::Undo()

   CASE nKey == xbeK_RIGHT .OR. nKey = xbeK_CTRL_D
        ::Get:Right()

   CASE nKey == xbeK_CTRL_RIGHT
        ::Get:WordRight()
        ::Get:clear := .f.

   CASE nKey == xbeK_LEFT .OR. nKey = xbeK_CTRL_S
        ::Get:Left()

   CASE nKey == xbeK_CTRL_LEFT
        ::Get:WordLeft()

   CASE nKey == xbeK_HOME
        ::Home()

   CASE nKey == xbeK_END
        ::Get:_End()

   CASE nKey == xbeK_INS
        Set( _SET_INSERT, ! Set( _SET_INSERT) )

   CASE nKey == xbeK_SH_INS .OR. nKey == xbeK_CTRL_V
        IF ::editable
           ::PasteMarked()
        ENDIF

   CASE nKey == xbeK_CTRL_INS .OR. nKey == xbeK_CTRL_C
        ::CopyMarked()

   CASE nKey == xbeK_BS

      IF ::isCalc

        ::xbpSle:keyBoard( nKey )
        cChr := StrTran( ::editBuffer(), ::thousand, "" )
        ::number := Val( StrTran( cChr, ::decimal , "." ) )

        IF ::number <> 0
          ::isMinus := ( ::number < 0 )
        ENDIF

        IF ::isFloat
          * Remove last character of the picture string.
          ::picture := SubStr( ::picture, 1, Len( ::picture )-1 )
          IF Right( ::picture, 1 ) == "."
            ::picture := SubStr( ::picture, 1, Len( ::picture )-1 )
          ENDIF
          ::isFloat := ( "." $ ::picture )

          IF ::isMinus .AND. ::number == 0
            * The edit buffer may contain something like "-0.00"
            ::isMinus := ::isFloat
          ENDIF
        ELSE
          * Remove first character of the picture string.
          ::picture := SubStr( ::picture, 2 )
          IF Left( ::picture, 1 ) == ","
            ::picture := SubStr( ::picture, 2 )
          ENDIF
          IF ::isMinus .AND. ::number == 0
            * Remove the minus sign when the edit buffer contains "-0."
            ::isMinus := .F.
          ENDIF
        ENDIF

        ::setEditbuffer()

      ELSE

        IF ::editable
           ::cutMarked()
           ::Get:BackSpace()
        ENDIF

      ENDIF

   CASE ::isCalc .AND. ::typeOut
      * The edit buffer is full.
      Tone(1000)

   CASE ::IsCalc .AND. cChr $ "0123456789"
      IF ::number == 0
         * The buffer may contain 0.00 or -0.00 when a number
         * between < 1 and and > -1 is entered.
         IF cChr == "0"
            IF .NOT. ::isFloat
               IF ::isMinus
                  ::picture := "9"
               ELSE
                  ::picture := ""
               ENDIF
            ENDIF
         ELSE
            IF ::isFloat
               ::number := Val( StrTran( ::editBuffer(), ::decimal, "." ) + cChr )
            ELSEIF ::isMinus
               ::number := - Val( cChr )
            ELSE
               ::number := Val( cChr )
            ENDIF
         ENDIF
      ELSE
         ::xbpSle:keyBoard( nKey )
         cChr := StrTran( ::editBuffer(), ::thousand, "" )
         ::number := Val( StrTran( cChr, ::decimal , "." ) )
      ENDIF

      DEFAULT ::picture := ''

      IF ::isFloat
         IF At( ".", ::picture ) == 0
            ::picture += "."
         ELSE
            ::picture += "9"
         ENDIF
      ELSE
         IF Left( ::picture, 3 ) == "999"
            ::picture := "9," + ::picture
         ELSE
            ::picture := "9" + ::picture
         ENDIF
      ENDIF
      ::setEditbuffer()

   CASE ::isCalc .AND. cChr $ ",."
      IF .NOT. ::isFloat
         ::isFloat := .T.
         IF ::number == 0
            ::picture += "9"
         ENDIF
         ::picture += "."
         ::setEditbuffer()
      ENDIF

   CASE ::isCalc .AND. cChr == "-"
      IF .NOT. ::isMinus
         ::isMinus := .T.
         ::number  *= -1
         IF Left( ::picture, 3 ) == "999"
            ::picture := "9," + ::picture
         ELSE
            ::picture := "9"  + ::picture
         ENDIF
         ::setEditbuffer()
      ENDIF

   CASE (nKey == xbeK_DEL .OR. nKey == xbeK_CTRL_G)
        IF ::editable
          IF Set(_SET_INSERT) .AND. aMarked[1] == aMarked[2]
             aMarked[2] := aMarked[1] + 1
             ::SetMarked(aMarked)
          ENDIF
          ::DeleteMarked()
        ENDIF

   CASE nKey == xbeK_SH_DEL .OR. nKey == xbeK_CTRL_X
        IF ::editable
           ::CutMarked()
        ENDIF

   CASE nKey == xbeK_CTRL_T
        IF ::editable
           ::Get:DelWordRight()
        ENDIF

   CASE nKey == xbeK_CTRL_O

   CASE nKey == xbeK_CTRL_Y
        IF ::editable
           IF ::Get:Type # 'D'
             ::Get:delEnd()
           ELSE
             nPos := ::Get:pos
             cPicture := ::Get:picture
             cSavePict := ::Get:picture
             IF Left(cPicture,1) = '@'
                cPicture := AllTrim(Substr(cPicture,At(' ',cPicture)+1))
             ENDIF
             ::Get:picture := ""
             FOR i := nPos TO ::bufferlength
               IF !(SubStr(cPicture,i,1) $ ' /.,')
                 ::Get:overstrike(' ')
               ELSE
                 ::Get:right()
               ENDIF
             NEXT
             ::Get:pos := nPos
             ::Get:picture := cSavePict
           ENDIF
        ENDIF

   CASE nKey == xbeK_CTRL_BS
        IF ::editable
           ::Get:DelWordLeft()
        ENDIF

   CASE nKey >= 32 .AND. nKey <= 255

//      cDecChar := ::decimal // SetLocale( NLS_SDECIMAL)

      cDecChar := SetLocale( NLS_SDECIMAL)

      IF ::get:type == 'D' .AND. ::dateFlag .AND. !Empty(::picture) .AND. 'K' $ Upper(::picture)
        ::get:VarPut(Ctod(''))
        ::setData()
        ::home()
      ELSEIF ::get:type == 'D' .AND. Chr(nKey) = '+'
        dDate := CtoD(::get:buffer)
        IF Empty(dDate)
          dDate := Date()-1
        ENDIF
        ::get:VarPut(++dDate)
        ::setData()
        ::home()
      ELSEIF ::get:type == 'D' .AND. Chr(nKey) = '-'
        dDate := CtoD(::get:buffer)
        IF Empty(dDate)
          dDate := Date()+1
        ENDIF
        ::get:VarPut(--dDate)
        ::setData()
        ::home()
      ELSEIF ::get:type = 'N' .AND. !::isCalc  .AND. ::get:pos > 1 .AND. Empty(::get:buffer[1])
        IF Valtype(::picture) = 'C' .AND. '@Z' $ ::picture
          nDecPos := At(cDecChar,Alltrim(Strtran(::picture,'@Z','')))
        ELSEIF Valtype(::picture) = 'C' .AND. '@$' $ ::picture
          nDecPos := At(cDecChar,Alltrim(Strtran(::picture,'@$','')))
        ELSE
          nDecPos := At(cDecChar,::get:Buffer)
        ENDIF
        IF ::get:pos < nDecPos .AND. Empty(::get:buffer[::get:pos+1])
          ::clear(.t.)
          ::home()
        ENDIF
      ENDIF

      nPos := ::Get:pos
      cChar := Chr(nKey)
      IF Valtype(::proper) == 'A'
        IF ::Get:pos=1 .OR. SUBSTR(::Get:buffer,::Get:pos-1,1)$::proper[2]
           cChar := Upper(cChar)
        ELSEIF ::proper[1]
           cChar := Lower(cChar)
        ENDIF
      ENDIF

      IF ::Get:Type == "N" .AND. cChar $ ".,"
         IF Val(Strtran(::get:buffer,'$','')) > 0 .AND. At( " ", Alltrim( ::Get:buffer)) != 0
//           ::Get:buffer := TransForm(Val(StrTran(Strtran(::Get:buffer,' ',''), ::thousand,'')),::Get:picture)
           ::Get:buffer := TransForm( Val( StrTran(Strtran(::Get:buffer,' ',''),',','')), ::Get:picture)
           ::SetEditBuffer(::Get:buffer)
         ENDIF

         ::get:ToDecPos()

/*
         IF ::thousand == ','
           ::get:ToDecPos()
         ELSE
           ::ToDecPos()
         ENDIF
*/
      ELSE
         IF ::editable
           IF ::Get:Type =="C" .AND. ::Picture!=nil .AND. ;
               ('@!' $ ::Picture .OR. Substr(::Picture,nPos,1) == '!' )
             IF (xPos := At(cChar,''))>0
               cChar := Substr('',xPos,1)
             ENDIF
           ENDIF
           IF Set(_SET_INSERT)
             IF Abs( aMarked[2]-aMarked[1]) > 0
               ::deleteMarked()
             ENDIF
             ::Get:Insert( cChar )
           ELSE
             IF Abs(aMarked[2]-aMarked[1]) > 1
               ::deleteMarked()
               ::Get:Insert( cChar )
             ELSE
               ::Get:Overstrike( cChar )
             ENDIF
          ENDIF
        ENDIF
      ENDIF

      IF !::confirm .AND. nPos = ::bufferLength .AND. ::get:changed
         PostAppEvent(xbeP_Keyboard,xbeK_ENTER,nil,self)
      ENDIF

   OTHERWISE

     lHandled := .F.

   ENDCASE

   IF ::get:type = 'D' .AND. ::dateFlag .AND. !Empty(::picture) .AND. 'K' $ Upper(::picture)
     ::dateFlag := .F.
   ENDIF

   IF !::isCalc
     IF (! ::EditBuffer() == ::Get:Buffer) .OR. ::password
        IF ::password
           ::passbuffer := ::Get:Buffer
          ::Get:Buffer := Pad(Repl(::passwordChar,Len(Alltrim(::passBuffer))),::bufferLength)
        ENDIF
        IF ::get:type = 'N' .AND. Empty(::get:buffer[::get:pos])
           cBuffer := Substr(::get:buffer,1,::get:pos)
           FOR i := ::get:pos+1 TO Len(::get:buffer)
             cBuffer += IIF(::get:buffer[i]$',.',::get:buffer[i],' ')
           NEXT
           ::get:buffer := cBuffer
        ENDIF
        ::SetEditBuffer( ::Get:Buffer )
        DC_GetEditUpdated(.t.)
     ENDIF

     IF setPos
        IF Set( _SET_INSERT )
          aMarked := { ::Get:Pos, ::Get:Pos }
        ELSE
          aMarked := { ::Get:Pos, ::Get:Pos + 1 }
       ENDIF
     ENDIF
     ::SetMarked( aMarked )

     IF Valtype(::keyBlock) = 'B' .AND. !Empty(nKey)
       Eval(::keyBlock,nKey,nil,self)
     ENDIF

     IF ! lHandled
       RETURN ::XbpSle:Keyboard( nKey )
     ENDIF
   ENDIF

RETURN self

* ------------

METHOD DC_XbpGet:ResizeSetSize( aSize )

LOCAL aArray, nWidth, nHeight, aGetListItem

aGetListItem := ::getList:getListArray[::getListPointer]

nWidth := ::currentSize()[1]
nHeight := ::currentSize()[2]

IF !Empty(::template)
  aArray := GraQueryTextBox( ::lockPS(), Repl(::template,Len(::Editbuffer())) )
ELSE
  aArray := GraQueryTextBox( ::lockPS(), ::EditBuffer() )
ENDIF

IF !Empty(aGetListItem[cGETLIST_FONT])
  IF !Empty(aGetListItem[nGETLIST_WIDTH])
    nWidth := aSize[1]
  ELSE
    nWidth := (aArray[3,1] - aArray[1,1]) + 6
  ENDIF
  nHeight := (aArray[1,2] - aArray[2,2]) + 4 + DC_XbpGetHeightPadding()
ELSE
  IF !Empty(aGetListItem[nGETLIST_WIDTH])
    nWidth := aSize[1]
  ELSE
    nWidth := Max(nWidth,(aArray[3,1] - aArray[1,1]) + 6)
  ENDIF
  nHeight := Max(nHeight,(aArray[1,2] - aArray[2,2]) + 2)
ENDIF

::unlockPS()

RETURN ::xbpSle:setSize( { nWidth, nHeight } )

* ------------

// Move the caret to the first editable character in the edit buffer.
METHOD DC_XbpGet:home()

LOCAL aMarked, lClear := ::get:clear

DEFAULT lClear := .f.

IF ::isCalc
  RETURN self
ENDIF

IF ::Get:hasFocus
  ::get:home()
ELSE
  ::Get:pos := 1
ENDIF

::get:clear := lClear .OR. ::get:type == 'N'
::XbpSle:setFirstChar( ::get:pos )

IF Set( _SET_INSERT )
  aMarked := { ::Get:Pos, ::Get:Pos }
ELSE
  aMarked := { ::Get:Pos, ::Get:Pos + 1 }
ENDIF

::SetMarked( aMarked )

RETURN self


* **********************************************************************
* Overloaded CutMarked()
* **********************************************************************
METHOD DC_XbpGet:CutMarked()

   ::CopyMarked()
   ::DeleteMarked()

RETURN self

* **********************************************************************
* Overloaded PasteMarked()
*
* J. Duijf 09-04-2006
* Had some complaints about the cut/paste behaviour in my applications.
* IMO, the new behaviour is more windows consistent, as in text editors.
*
* The paste behaviour is now as follows:
*
* First all marked characters are deleted.
* Then the paste buffer is _INSERTED_  Regardless the _SET_INSERT mode
*
* **********************************************************************
METHOD DC_XbpGet:PasteMarked()

   LOCAL cString, oClipboard, i
   LOCAL lInsert  := SET(_SET_INSERT)                                         // Current insert mode
   LOCAL aMarked := ::QueryMarked()                                           // Current marked area

   oClipboard := XbpClipboard():New():Create()
   oClipboard:Open()
   cString := oClipboard:GetBuffer( XBPCLPBRD_TEXT )                          // get string from clipboard
   oClipboard:Close()

   if !lInsert
      //
      // There is always 1 character marked when in overstrike mode. Do not overwrite
      //
      if aMarked[1]+1 = aMarked[2]                                            // Only 1 character is marked
         ::SetMarked({aMarked[1],aMarked[1]})                                 // Unmark area
      else
         // More then 1 character is marked, proceed.
      endif
   endif
   ::DeleteMarked()                                                           // Delete marked area (if any)

   * paste an empty string into the edit buffer when the clipboard
   * does not contain a valid string


   IF cString == NIL ; cString := "" ; endif
   For i := 0 To 31           ; cString := Strtran(cString,Chr(i),''); Next i // Remove ASCII controll characters (if any)
   For i := 1 To Len(cString) ; ::Get:Insert(Substr(cString,i,1))    ; Next i // Insert into edit buffer
   *
   * Update Dc_Get
   *

   IF ! ::EditBuffer() == ::Get:Buffer
      ::SetEditBuffer( ::Get:Buffer )
      DC_GetEditUpdated(.t.)
   ENDIF

RETURN self

/*
 * Overloaded :deleteMarked() method
 */
METHOD DC_XbpGet:DeleteMarked()

   LOCAL aMarked := ::QueryMarked(), i
   LOCAL nMarked := Abs( aMarked[2] - aMarked[1] )

   IF nMarked == 0
       // We're in insert mode and nothing is marked -> delete nothing.
   ELSEIF nMarked == 1
      ::Get:Delete()
   ELSE
      ::Get:Pos := Max( aMarked[1], aMarked[2] )
      FOR i := 1 TO nMarked
         ::Get:BackSpace()
      NEXT
   ENDIF
   IF ! ::EditBuffer() == ::Get:Buffer
      IF !::password
        ::SetEditBuffer( ::Get:Buffer )
      ENDIF
      DC_GetEditUpdated(.t.)
   ENDIF

RETURN self

/* ------------------ */

METHOD DC_XbpGet:GetSetFocus()

  LOCAL cBuffer, nBuffer

  IF ::Get:Type == 'N'
     cBuffer := ::Get:Buffer
     nBuffer := VAL(StrTran(cBuffer, ',', '') )
     ::Get:Picture := PADL(StrTran(::Get:Picture, ',', ''), LEN(::Picture))
     cBuffer := TRANSFORM(nBuffer, ::Get:Picture)
     ::Get:Buffer := cBuffer
     ::SetEditBuffer (cBuffer)
  ENDIF

RETURN self

* ------------------------

METHOD DC_XbpGet:GetKillFocus()

  LOCAL cBuffer

  IF ::Get:Type == 'N'
     IF At( " ", Alltrim( ::Get:buffer)) != 0
       ::Get:buffer := TransForm( Val( ::Get:buffer), ::Get:picture)
     ENDIF
     cBuffer := ::Get:Buffer
     cBuffer := TRANSFORM(VAL(cBuffer), ::Picture)
     ::Get:Picture := ::Picture
     ::SetEditBuffer (cBuffer)
  ENDIF

RETURN self

* ---------------

METHOD DC_XbpGet:hideShow()

RETURN ::WhenHideEval()

* ---------------

METHOD DC_XbpGet:WhenHideEval( lEnable, lShow, lProtect )

LOCAL oXbpPick := ::popupButton
LOCAL oSay := ::sayObject
LOCAL aCaption, xCaption1, xCaption2, lPopEnable, lPopHide

IF Valtype(::whenBlock) = 'B'
  lEnable:= Eval(::whenBlock,self)
  IF lEnable
    IF !::isEnabled()
      ::enable()
      IF Valtype(oXbpPick) = 'O' // enable the POPUP button
        aCaption := oXbpPick:cargo[2]
        IF Valtype(aCaption) = 'A'
          xCaption1 := aCaption[1]
          IF Valtype(xCaption1) = 'B'
            xCaption1 := Eval(xCaption1)
          ENDIF
          oXbpPick:setCaption(xCaption1)
        ELSE
          oXbpPick:enable()
        ENDIF
        oXbpPick:cargo[3] := .t.
      ENDIF
    ENDIF
  ELSE
    IF ::isEnabled()
      ::disable()
      IF Valtype(oXbpPick) = 'O' // disable the POPUP button
        aCaption := oXbpPick:cargo[2]
        IF Valtype(aCaption) = 'A'
          xCaption2 := aCaption[2]
          IF Valtype(xCaption2) = 'B'
            xCaption2 := Eval(xCaption2)
          ENDIF
          oXbpPick:setCaption(xCaption2)
        ELSE
          oXbpPick:disable()
        ENDIF
        oXbpPick:cargo[3] := .f.
      ENDIF
    ENDIF
  ENDIF
ENDIF

IF Valtype(::protectBlock) == 'B' .AND. (Valtype(lProtect) = 'L' .OR. DC_GetPopupProtectHide())
  lProtect := Eval(::protectBlock,self)
  IF Valtype(oXbpPick) == 'O' .AND. DC_GetPopupProtectHide()
    IF lProtect
      oXbpPick:hide()
    ELSEIF Valtype(lShow)=='L' .AND. lShow
      oXbpPick:show()
    ENDIF
  ENDIF
ENDIF

IF Valtype(::hideBlock) = 'B'
  lShow := !Eval(::hideBlock,self)
  IF lShow
    ::show()
    IF Valtype(oXbpPick) = 'O' .AND. ;
       (Valtype(lProtect) # 'L' .OR. !lProtect .OR. !DC_GetPopupProtectHide())
      IF !DC_GetPopupAutoHide() .OR. SetAppFocus()==self
        oXbpPick:show()
      ENDIF
    ENDIF
    IF Valtype(oSay) = 'O'
      oSay:show()
    ENDIF
  ELSE
    ::hide()
    IF Valtype(oXbpPick) = 'O'
      oXbpPick:hide()
    ENDIF
    IF Valtype(oSay) = 'O'
      oSay:hide()
    ENDIF
  ENDIF
ENDIF

IF Valtype(::popupWhen) == 'B'
  lPopEnable = Eval(::popupWhen,self)
  IF !lPopEnable .AND. ::popupButton:isEnabled()
    ::popupButton:disable()
  ELSEIF lPopEnable .AND. !::popupButton:isEnabled()
    ::popupButton:enable()
  ENDIF
ENDIF

IF Valtype(::popupHide) == 'B'
  lPopHide = Eval(::popupHide,self)
  IF lPopHide .AND. ::popupButton:isVisible()
    ::popupButton:hide()
  ELSEIF !lPopHide .AND. !::popupButton:isVisible()
    ::popupButton:show()
  ENDIF
ENDIF

RETURN self

* ------------

METHOD DC_XbpGet:makePicture()

LOCAL cBuffer, nLen, nDec

   IF ::number == 0
      ::picture := ""
      ::isMinus := .F.
      ::isFloat := .F.
      ::typeOut := .F.
   ELSE
      ::isMinus := ( ::number < 0 )

      cBuffer := LTrim( Str( ::number, ::bufferLength+32, ::bufferLength ) )
      nLen    := Len( cBuffer )
      DO WHILE SubStr( cBuffer, nLen, 1 ) == "0"
         nLen --
      ENDDO

      cBuffer := SubStr( cBuffer, 1, nLen )
      IF Right( cBuffer, 1 ) == "."
         nLen --
         cBuffer := SubStr( cBuffer, 1, nLen )
      ENDIF

      nDec := At( ".", cBuffer )
      IF nDec > 0
         ::isFloat := .T.
         ::picture := "." + Replicate( "9", nLen-nDec )
         nLen := nDec - 1
         nDec := 0
      ELSE
         ::isFloat := .F.
      ENDIF

      DEFAULT ::picture := ''
      DO WHILE nLen > 0
         ::picture := "9" + ::picture
         nDec ++
         nLen --
         IF nDec == 3
            nDec := 0
            IF nLen > 0
              ::picture := "," + ::picture
            ENDIF
         ENDIF
      ENDDO
   ENDIF

RETURN self


* Determine from the font size if the visible part of the entry
* field is entirely filled with digits.

METHOD DC_XbpGet:typeOut

   LOCAL oPS        := ::lockPS()
   LOCAL aBox1      := GraQueryTextBox( oPS, "9" )
   LOCAL nFontWidth := aBox1[3,1] - aBox1[1,1]
   LOCAL aBox2      := GraQueryTextBox( oPS, ::editBuffer() )
   LOCAL nBuffWidth := aBox2[3,1] - aBox2[1,1]
   LOCAL nDispWidth := ::currentSize()[1] - 2 * nFontWidth

   ::unlockPS(oPS)
   ::typeOut := ( nBuffWidth > nDispWidth )

   IF ::typeOut
      ::xbpSle:typeOut()
   ENDIF

RETURN self


// John Lau: New class added for support of Chinese char. input in XbpGet.
* **********************************************************************
* CXbpGet class
* **********************************************************************

CLASS DC_XbpGetC FROM DC_XbpGet

   HIDDEN:

   EXPORTED:

   VAR    nTrueBufferLen          // true buffer length
   VAR    lPassThru               // when .T. let XbpGet handle everything
                                  ** overloaded from XbpGet and XbpSle
   METHOD Create                  // Lifecycle
   METHOD Keyboard                // overloaded keyboard handling
   METHOD LbUp                    // to inquire mark
   METHOD RbUp                    // to fix cut and paste bug

   METHOD At_half_DBCS

ENDCLASS

* **********************************************************************
* Create method
* **********************************************************************
METHOD DC_XbpGetC:Create( oParent, oOwner, aPos, aSize, aPP, lVisible, ;
                          aGetList, nGetPointer, aOptions )

   * create XbpGet first
   ::DC_XbpGet:Create( oParent, oOwner, aPos, aSize, aPP, lVisible, ;
                       aGetList, nGetPointer, aOptions )

   * now invisible GET should also have been initialized

   * now the datalink must have been specified
   IF ! ValType( ::DataLink ) == "B"
      ::Status := XBP_STAT_FAILURE
      RETURN self
   ENDIF

   ::nTrueBufferLen  := Len( ::Get:buffer )

   // This may not be needed since it has been set to 1000 by eXPress++ anyways
   ::BufferLength    := ::nTrueBufferLen * 2

   ::lPassThru := (!::type $ 'C' .OR. !empty(::Picture) .OR. ::password)

   // Not required anymore in Xbase++ 1.3,
   // - but the need special XPPUI1.DLL from Alaska.
   *IF ! ::lPassThru
   *   ::XbpSle:SetEventMask(2, .T.)    // the important statement to allow
   *                                    // DBCS input to be active.
   *ENDIF

RETURN self


* **********************************************************************
* Overloaded Keyboard handling
* **********************************************************************
METHOD DC_XbpGetC:Keyboard( nKey )

   LOCAL aMarked, lHotKeyDone := .F.
   LOCAL lsetPos := .T., xRtnVal := self
   Local EditBuffer
   IF ::lPassThru
      RETURN ::XbpGet:Keyboard( nKey )
   ENDIF

   * --- This must be tested first ---
   * --- comment out this case if we want BS to be W32 compatiable ---
   if nKey >= 32 .AND. nKey <= 255
        return ::XbpGet:Keyboard( nKey )
   endif

   sleep(2)    // added for Xbase++ 1.3
   aMarked := ::QueryMarked()
   EditBuffer := ::EditBuffer()

   DO CASE

   CASE nKey == xbeK_BS
        * --  the following is XbpGet compatible, and handles DBCS --
        IF ::editable
            IF ::Get:Pos > 1 .and. ::At_half_DBCS(::Get:Buffer, ::Get:Pos-1)
               ::Get:BackSpace()
            endif
            ::Get:BackSpace()
            ::SetEditBuffer( ::Get:Buffer )
            DC_GetEditUpdated(.t.)
            * lsetPos := .T.   // default is .T.
        else
            lsetPos := .F.
        ENDIF

   * --- This must be tested next ---
   CASE ! trim(EditBuffer) == trim(::Get:Buffer)
        * update Get:Buffer from SLE's EditBuffer()
        ::Get:Buffer := pad(EditBuffer, ::nTrueBufferLen)

        * move cursor position in GET object (could be up or down)
        ::Get:pos := aMarked[1]
        IF ::Get:pos >  ::nTrueBufferLen  // Get:Pos always <= nTrueBufferLen
           ::Get:pos := ::nTrueBufferLen
        ENDIF
        IF ::At_half_DBCS(::Get:Buffer, ::nTrueBufferLen )
           * last char. of Get:Buffer is at half DBCS, chop it off.
           ::Get:Buffer := Left(::Get:Buffer, ::nTrueBufferLen - 1) + ' '
        ENDIF

        IF !::confirm .AND. ::Get:pos == ::nTrueBufferLen
           PostAppEvent(xbeP_Keyboard,xbeK_ENTER,nil,self)
        ENDIF
        * Finally synchronise buffer from GET object.
        ::SetEditBuffer( ::Get:Buffer )
        DC_GetEditUpdated(.t.)

        * lsetPos := .T.   // default is .T.

   CASE nKey == xbeK_SH_RIGHT
        ::Get:Pos ++
        IF ::At_half_DBCS(::Get:Buffer, ::Get:Pos)
           ::Get:Pos ++
        ENDIF
        aMarked[2] := ::Get:Pos
        ::SetMarked( aMarked )
        lsetPos := .F.

   CASE nKey == xbeK_SH_LEFT
        if ::Get:Pos > 1
           aMarked[2] ++
           IF ::At_half_DBCS(::Get:Buffer, aMarked[2])
              aMarked[2] ++
           ENDIF

           ::Get:Pos --
           IF ::Get:Pos > 1 .and. ::At_half_DBCS(::Get:Buffer, ::Get:Pos)
              ::Get:Pos --
           ENDIF
           aMarked[1] := ::Get:Pos
           ::SetMarked( aMarked )
           lsetPos := .F.
        endif

   CASE nKey == xbeK_SH_END
        aMarked[2] := ::nTrueBufferlen + 1
        ::Get:Pos  := ::nTrueBufferlen
        ::SetMarked( aMarked )
        lsetPos := .F.

   CASE nKey == xbeK_RIGHT .OR. nKey = xbeK_CTRL_D
        ::Get:Right()
        IF ::At_half_DBCS(::Get:Buffer, ::Get:Pos)
           ::Get:Right()
        ENDIF
        * lsetPos := .T.   // default .T.

   CASE nKey == xbeK_LEFT .OR. nKey = xbeK_CTRL_S
        ::Get:Left()
        IF ::At_half_DBCS(::Get:Buffer, ::Get:Pos)
           ::Get:Left()
        ENDIF
        * lsetPos := .T.   // default .T.

   CASE (nKey == xbeK_DEL        .OR. ;
         nKey == xbeK_CTRL_G)    .OR. ;
         nKey == xbeK_SH_HOME    .OR. ;
         nKey == xbeK_CTRL_U     .OR. ;
         nKey == xbeK_CTRL_RIGHT .OR. ;
         nKey == xbeK_CTRL_LEFT  .OR. ;
         nKey == xbeK_HOME       .OR. ;
         nKey == xbeK_END        .OR. ;
         nKey == xbeK_INS        .OR. ;
         nKey == xbeK_SH_INS     .OR. ;
         nKey == xbeK_CTRL_INS   .OR. ;
         nKey == xbeK_SH_DEL     .OR. ;
         nKey == xbeK_CTRL_T     .OR. ;
         nKey == xbeK_CTRL_Y     .OR. ;
         nKey == xbeK_CTRL_BS

        * --- do these in XbpGet compatible way ---
        xRtnVal := ::XbpGet:Keyboard( nKey )
        lsetPos := .F.
        lHotKeyDone := .T.

   OTHERWISE
        xRtnVal := ::XbpSLE:Keyboard( nKey )
        lSetPos  := .F.
   ENDCASE

   IF ! lHotKeyDone
      IF Valtype(::keyBlock) = 'B' .AND. !Empty(nKey)
         Eval(::keyBlock,nKey,nil,self)
      ENDIF
   ENDIF

   * get value of aMarked again...
   IF lSetPos
      * buffer handled here, obtain aMarked from Get:Pos
      IF Set( _SET_INSERT )
         aMarked := { ::Get:Pos, ::Get:Pos }
      ELSE
         aMarked := { ::Get:Pos, ::Get:Pos + 1}
      ENDIF
   ELSE
      * get what's been set by XbpGet:Keyboard or other set above
      aMarked := ::QueryMarked()
   ENDIF

   * ... and re-set aMarked after checking for DBCS boundary.
   IF ::At_half_DBCS(::Get:Buffer, aMarked[1])
      aMarked[1] := aMarked[1] - 1
   ENDIF
   IF ::At_half_DBCS(::Get:Buffer, aMarked[2])
      aMarked[2] := aMarked[2] + 1
   ENDIF

   ::setmarked(aMarked)

RETURN xRtnVal


* **********************************************************************
* Overloaded LbUp
* **********************************************************************
METHOD DC_XbpGetC:LbUp(mp1, mp2)

   LOCAL aMarked := ::queryMarked()

   * now locate Get on the specified position
   IF ::At_half_DBCS(::Get:Buffer, aMarked[1])
      aMarked[1] --
   ENDIF
   ::Get:Pos := aMarked[1]

   * now set the mark when insert mode is not active
   IF aMarked[2] - aMarked[1] == 0 .AND. ! Set( _SET_INSERT )
      IF ::At_half_DBCS(::Get:Buffer, aMarked[1] + 1)
         ::SetMarked( { aMarked[1], aMarked[1] + 2 } )
      ELSE
         ::SetMarked( { aMarked[1], aMarked[1] + 1 } )
      ENDIF
   ENDIF

RETURN ::XbpSle:LbUp( mp1, mp2 )
//RETURN ::XbpGet:LbUp( mp1, mp2 )

/* ----------------- */

METHOD DC_XbpGetC:RBUp(aPos)

::Get:Buffer := pad(::EditBuffer(), ::nTrueBufferLen)

RETURN ::XbpSle:RbUp(aPos)
/////


/* ----------------- */
* **********************************************************************
* At_half_DBCS (cBuffer, nPos)
* - check if nPos char. of cBuffer is 2nd half of DBCS char.
* **********************************************************************
METHOD DC_XbpGetC:At_half_DBCS(cBuffer, nPos)

  LOCAL lHalf_DBCS := .F.     // default no.
  LOCAL nCNT
  LOCAL cChar, nChar

  FOR nCNT := 1 to LEN(cBuffer)
      IF nCNT == nPos
         exit
      ENDIF
      IF lHalf_DBCS
         lHalf_DBCS := .F.
      ELSE
         cChar := substr(cBuffer, nCNT, 1)
         nChar := asc(cChar)
         IF nChar > 127
            lHalf_DBCS := .T.
         ENDIF
      ENDIF
  NEXT

RETURN lHalf_DBCS

* ------------

STATIC FUNCTION _PopKey( oGet )

RETURN {||SetAppFocus(oGet),Eval(oGet:popupButton:activate)}

* ------------------ *

STATIC FUNCTION _GetCombo( oGet, lPixel, aOptions, nStyle )

LOCAL nHeight, xData, oParent, bField, nCol, nRow, nWidth, ;
      nPointer := 1, bReturn, cFont, aPres

DEFAULT nStyle := 0

IF lPixel
  nHeight := oGet:comboHeight
ELSE
  nHeight := oGet:comboHeight * aOptions[nGETOPT_ROWPIXELS]
ENDIF

xData   := oGet:comboData
bField  := oGet:comboField
bReturn := oGet:comboReturn

IF lPixel
  nWidth  := oGet:comboWidth
ELSEIF !Empty(oGet:comboWidth)
  nWidth  := oGet:comboWidth * aOptions[nGETOPT_COLPIXELS]
ENDIF
DEFAULT nWidth := oGet:currentSize()[1]+4

cFont := oGet:comboFont
aPres := oGet:comboPres

IF bField == NIL
  bField := 1
ENDIF
oParent := oGet:setParent()
nCol    := oGet:currentPos()[1]
nRow    := oGet:currentPos()[2]

oParent := AppDeskTop()

oGet:comboStatic := XbpDialog():new( oParent, oParent, { nCol, nRow-nHeight }, ;
                       {nWidth+IIF(nStyle=0,24,0),nHeight},,.f.)

oGet:comboStatic:titleBar := .f.
oGet:comboStatic:clipSiblings := .t.
oGet:comboStatic:create()
oGet:comboStatic:hide()
oGet:comboStatic:drawingArea:setFontCompoundName(oGet:setFontCompoundName())

RETURN {|c,o,nKey|_GetComboData(c,bField,nHeight,xData, ;
                         @nPointer,bReturn,nWidth,@oGet,cFont,aPres,nKey) }

* -----------------------

STATIC FUNCTION ;
  _GetComboData( c, bField, nHeight, xData, nPointer, bReturn, nWidth, ;
                 oGet, cFont, aPres, nKey)

LOCAL nLength, lOk, acbData, i, aPos, aWorkArea, aGetPos, bItemMarked

IF Valtype(c) = 'C'
  nLength := Len(c)
ELSE
  nLength := Len(oGet:editBuffer())
ENDIF

xData := oGet:comboData

IF Valtype(xData) = 'A'
  IF !Empty(xData) .AND. Valtype(xData[1]) # 'A'
    acbData := Array(Len(xData))
    FOR i := 1 TO Len(xData)
      acbData[i] := { xData[i] }
    NEXT
  ELSE
    acbData := xData
  ENDIF
ELSE
  acbData := xData
ENDIF

aGetPos := DC_CalcAbsolutePosition( {0,0}, oGet )
aPos := { aGetPos[1]+3, aGetPos[2]-nHeight+3 }

aWorkArea := DC_GetWorkArea()

IF aPos[2] < ( AppDeskTop():currentSize()[2] - aWorkArea[4] )
  aPos[2] := aGetPos[2] + oGet:currentSize()[2]
ENDIF

oGet:comboStatic:setPos(aPos)

oGet:comboStatic:show()
oGet:comboStatic:toFront()

IF DC_GetComboImmediate()
  bItemMarked := {|a,b,o,c|c:=_GetComboReturn(acbData,bField,bReturn,@nPointer,nLength,oGet)}
ENDIF

nPointer := 1

lOk := DC_FindBrowse( {{bField,'',nWidth,nil,nil}}, ;
                       oGet:comboStatic:drawingArea, 2, 2, ;
                       oGet:comboStatic:currentSize()[1]-9, nHeight-9, ;
                       nil, .f., .f., .t., acbData, @nPointer,,aPres,,,,,cFont, ;
                       nil, nil, nil, nil, nil, bItemMarked,, nKey, oGet)

oGet:comboStatic:hide()

IF lOk
  c := _GetComboReturn(acbData,bField,bReturn,nPointer,nLength,IIF(!Empty(nKey),oGet,nil))
ENDIF

RETURN c

* -----------------

STATIC FUNCTION _GetComboReturn(acbData,bField,bReturn,nPointer,nLength,oGet)

LOCAL c

IF Valtype(bReturn) = 'B'
  c := Eval(bReturn,nPointer)
ELSEIF Valtype(acbData) = 'C'
  c := Eval(bField,nPointer)
ELSEIF Valtype(acbData) $ 'A' .AND. Valtype(bField) = 'N'
  c := acbData[nPointer,bField]
ELSEIF Valtype(acbData) $ 'B' .AND. Valtype(bField) = 'N'
  c := Eval(acbData)[nPointer,bField]
ENDIF

IF Valtype(c) = 'C'
  c := Pad(c,nLength)
ENDIF

IF Valtype(oGet) = 'O'
  oGet:_Assign(c)
ENDIF

RETURN c

* -----------------

STATIC FUNCTION _GetComboKeyboard( bCombo, nKey, oGet )

LOCAL cBuffer := oGet:editBuffer()

IF !Empty(cBuffer) .AND. Len(Alltrim(cBuffer)) == 1 ;
   .AND. nKey > 31 .and. nKey < 254 .AND. !oGet:comboFlag
  Eval(bCombo, nil, nil, xbeK_END )
  SetAppFocus(oGet)
  oGet:comboFlag := .T.
ENDIF

RETURN nil

* -----------------

STATIC FUNCTION _PopClick( bPopUp, oGet, bRef, lPopValid, getList )

RETURN {|x|x:=_ReadClick(bPopUp,oGet,bRef,lPopValid,getList),x }

* ----------------

STATIC FUNCTION _ReadClick ( bEval, oGet, xRef, lPopValid, getList )

LOCAL oAppFocus := oGet, xValue, lLocked

lPopValid := IIF(Valtype(lPopValid)='L',lPopValid,.f. )

xValue := Eval(oGet:datalink,oGet:GetData(),oGet,xRef)

oGet:exitState := MOUSE_ENTER
oGet:get:exitState := MOUSE_ENTER

oGet:popupButton:disable()
IF lPopValid
  xValue := Eval( bEval, oGet )
ELSE
  xValue := Eval( bEval, xValue, oGet, xRef )
ENDIF
oGet:popupButton:enable()

IF !lPopValid
  IF Valtype(xValue) = 'A'
    oAppFocus := xValue[2]
    xValue := xValue[1]
  ENDIF
  IF Valtype(xValue) == 'C'
    xValue := Pad(xValue,oGet:bufferLength)
  ENDIF
  IF oGet:password
    oGet:_assign( Repl(oGet:passwordChar,Len(Alltrim(xValue))) )
    oGet:passbuffer := xValue
  ELSE
    oGet:_assign(xValue)
  ENDIF
ENDIF
getList:lastFocus := oGet
IF !(oAppFocus==oGet)
  SetAppFocus(oGet)
ENDIF
SetAppFocus(oAppFocus)

RETURN xValue

* ---------------------

STATIC FUNCTION _PopUpButtonActivate( a, b, oButton )

IF Valtype(oButton:cargo[4]) == 'B' .AND. oButton:cargo[3]
  IF Empty(oButton:cargo[5]) .OR. ;
      (Valtype(oButton:cargo[5])=='B' .AND. !Eval(oButton:cargo[5]) ) // is EDITPROTECTed
     Eval( oButton:cargo[4], a, b, oButton )
     #if XPPVER > 1900000
       PostAppEvent( xbeM_Leave,,,oButton )
     #endif
  ENDIF
ENDIF

RETURN nil

* ----------------------

FUNCTION DC_ColorGets( o, nMode, aColor )

LOCAL aPres

IF !Empty(aColor) .AND. Empty(o:color)

  IF Valtype(aColor) == 'B'
    aColor := Eval(aColor)
  ENDIF
  IF Empty(aColor)
    RETURN nil
  ENDIF
  IF Len(aColor) < 2
    AAdd(aColor,{ GRA_CLR_BLACK, GRA_CLR_WHITE })
  ENDIF
  IF nMode = 1
    aPres := { { XBP_PP_FGCLR, aColor[1,1] } , { XBP_PP_BGCLR, aColor[1,2] } }
  ELSE
    aPres := { { XBP_PP_FGCLR, aColor[2,1] } , { XBP_PP_BGCLR, aColor[2,2] } }
  ENDIF
  o:setPresParam(aPres)
ENDIF
RETURN nil

* ----------------------

STATIC FUNCTION _AddHotKey( nAccel, bAction, bWhen, bHide, oObject, aKeyboard, oXbp )

LOCAL oFocusObject

IF Valtype(bAction) # 'B' .AND. Valtype(oObject) = 'O'
  bAction := {||SetAppFocus(oObject)}
  oFocusObject := nil
ELSE
  oFocusObject := oObject
ENDIF

IF Valtype(bAction) = 'B'
  AAdd( aKeyboard, { nAccel, bAction, bWhen, bHide, oFocusObject, oXbp, nil, .T. } )
ENDIF

RETURN nil

* ------------------

FUNCTION DC_GetPopupCaption( aCaption )

STATIC saCaption := ''
LOCAL aOldCaption
IF Empty(saCaption)
  saCaption := ;
    { '12.Arial Bold', ;
       Repl(Chr(249),3), ;
       0, ;
       nil, ;
       nil, ;
       nil }
ENDIF
aOldCaption := saCaption
IF Valtype(aCaption) = 'A'
  ASize(aCaption,6)
  saCaption := aCaption
ENDIF
RETURN aOldCaption

* ------------------

STATIC FUNCTION _PopupButtonShow( oGet, lShow )

IF !DC_GetPopupAutoHide() .OR. Empty(oGet:popupButton)
  RETURN nil
ENDIF
IF lShow
  oGet:popupButton:show()
ELSEIF SetAppFocus() # oGet:popupButton
  oGet:popupButton:hide()
ENDIF

RETURN nil

* ------------------

GETSETFUNCTION DC_GetPopupProtectHide DEFAULT .f.
GETSETFUNCTION DC_GetPopupAutoHide DEFAULT .f.

* ------------------

FUNCTION DC_GetReadInsert( nMode )

STATIC snMode := 0
LOCAL nOldMode := snMode

IF Valtype(nMode) = 'L'
  IF nMode
    nMode := 2 // Insert Mode
  ELSE
    nMode := 1 // Overstrike Mode
  ENDIF
ENDIF

IF Pcount() = 1
  snMode := nMode
ENDIF

RETURN nOldMode

* ------------------

FUNCTION DC_XbpGetPres( aPres )

STATIC saPres

LOCAL aOldPres := saPres

IF Valtype(aPres) = 'A'
  saPres := aPres
ENDIF

RETURN aOldPres

* -----------------

GETSETFUNCTION DC_XbpGetHeightPadding DEFAULT 0

* -----------------

GETSETFUNCTION DC_XbpGetLbDown DEFAULT {|x,y,o|o:Clear(.f.)}

* -----------------

GETSETFUNCTION DC_XbpGetLbClick DEFAULT {||nil}



